# Copyright 2022-2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import re
import shutil
import pathlib
from datetime import datetime
from dataclasses import dataclass
from typing import Optional, List, Dict, Tuple, Union

from . import github
from .families import Family

# -- Assorted Helper Functions
ISO8601_FMT = "%Y-%m-%dT%H:%M:%SZ"


def date_to_iso8601(date: datetime) -> str:
    return date.strftime(ISO8601_FMT)


def date_from_iso8601(string: str) -> datetime:
    return datetime.strptime(string, ISO8601_FMT)


def mkdirp(path):
    return pathlib.Path(path).mkdir(parents=True, exist_ok=True)


# -- API Variables

# -- PDK Root Management
VOLARE_DEFAULT_HOME = os.path.join(os.path.expanduser("~"), ".volare")
VOLARE_RESOLVED_HOME = os.getenv("PDK_ROOT") or VOLARE_DEFAULT_HOME


def _get_current_version(pdk_root: str, pdk: str) -> Optional[str]:
    current_file = os.path.join(get_volare_dir(pdk_root, pdk), "current")
    current_file_dir = os.path.dirname(current_file)
    mkdirp(current_file_dir)
    version = None
    try:
        version = open(current_file).read().strip()
    except FileNotFoundError:
        pass

    return version


def get_volare_home(pdk_root: Optional[str] = None) -> str:
    return pdk_root or VOLARE_RESOLVED_HOME


def get_volare_dir(pdk_root: str, pdk: str) -> str:
    return os.path.join(pdk_root, "volare", pdk)


def get_versions_dir(pdk_root: str, pdk: str) -> str:
    return os.path.join(get_volare_dir(pdk_root, pdk), "versions")


@dataclass
class Version(object):
    name: str
    pdk: str
    commit_date: Optional[datetime] = None
    upload_date: Optional[datetime] = None
    prerelease: bool = False

    def __lt__(self, rhs: "Version"):
        return (self.commit_date or datetime.min) < (rhs.commit_date or datetime.min)

    def __str__(self) -> str:
        return self.name

    def is_installed(self, pdk_root: str) -> bool:
        version_dir = get_version_dir(pdk_root, self.pdk, self.name)
        return os.path.isdir(version_dir)

    def is_current(self, pdk_root: str) -> bool:
        return self.name == _get_current_version(pdk_root, self.pdk)

    def get_dir(self, pdk_root: str) -> str:
        return os.path.join(get_versions_dir(pdk_root, self.pdk), self.name)

    def unset_current(self, pdk_root: str):
        if not self.is_installed(pdk_root):
            return
        if not self.is_current(pdk_root):
            return

        for variant in Family.by_name[self.pdk].variants:
            try:
                os.unlink(os.path.join(pdk_root, variant))
            except FileNotFoundError:
                pass

        current_file = os.path.join(get_volare_dir(pdk_root, self.pdk), "current")
        os.unlink(current_file)

    def uninstall(self, pdk_root: str):
        if not self.is_installed(pdk_root):
            raise ValueError(
                f"Version {self.name} of the {self.pdk} PDK is not installed."
            )

        self.unset_current(pdk_root)

        version_dir = self.get_dir(pdk_root)

        shutil.rmtree(version_dir)

    @classmethod
    def get_current(Self, pdk_root: str, pdk: str) -> Optional["Version"]:
        current_version = _get_current_version(pdk_root, pdk)
        if current_version is None:
            return None

        return Version(current_version, pdk)

    @classmethod
    def get_all_installed(Self, pdk_root: str, pdk: str) -> List["Version"]:
        versions_dir = get_versions_dir(pdk_root, pdk)
        mkdirp(versions_dir)
        return [
            Version(
                name=version,
                pdk=pdk,
            )
            for version in os.listdir(versions_dir)
            if os.path.isdir(os.path.join(versions_dir, version))
        ]

    @classmethod
    def _from_github(Self) -> Dict[str, List["Version"]]:
        releases = github.get_releases()

        rvs_by_pdk: Dict[str, List["Version"]] = {}

        commit_rx = re.compile(r"released on ([\d\-\:TZ]+)")

        for release in releases:
            if release["draft"]:
                continue
            family, hash = release["tag_name"].split("-")

            upload_date = date_from_iso8601(release["published_at"])
            commit_date = None

            commit_date_match = commit_rx.search(release["body"])
            if commit_date_match is not None:
                commit_date = date_from_iso8601(commit_date_match[1])

            remote_version = Self(
                name=hash,
                pdk=family,
                commit_date=commit_date,
                upload_date=upload_date,
                prerelease=release["prerelease"],
            )

            if rvs_by_pdk.get(family) is None:
                rvs_by_pdk[family] = rvs_by_pdk.get(family) or []

            rvs_by_pdk[family].append(remote_version)

        for family in rvs_by_pdk.keys():
            rvs_by_pdk[family].sort(reverse=True)

        return rvs_by_pdk

    def get_release_links(
        self, scl_filter: Optional[List[str]] = None
    ) -> List[Tuple[str, str]]:
        default_filter = False
        if scl_filter is None:
            default_filter = True
            scl_filter = Family.by_name[self.pdk].default_includes

        release = github.get_release_links(f"{self.pdk}-{self.name}")

        assets = release["assets"]
        zst_files = []
        xz_file = None
        for asset in assets:
            if default_filter and asset["name"] == "default.tar.xz":
                xz_file = (asset["name"], asset["browser_download_url"])
            elif asset["name"].endswith(".tar.zst"):
                asset_scl = asset["name"][:-8]
                if (
                    asset_scl == "common"
                    or "all" in scl_filter
                    or asset_scl in scl_filter
                ):
                    zst_files.append((asset["name"], asset["browser_download_url"]))

        if len(zst_files):
            return zst_files
        if xz_file is not None:
            return [xz_file]
        if scl_filter is not None:
            raise ValueError(
                f"No files found for standard cell libraries: {scl_filter}."
            )

        raise Exception(
            f"The release for {self.pdk}-{self.name} is malformed. Please file a bug report."
        )


def resolve_version(
    version: Optional[str],
    tool_metadata_file_path: Optional[str],
) -> str:
    """
    Takes an optional version and tool_metadata_file_path.

    If version is set, it is returned.

    If not, tool_metadata_file_path is checked if it exists.

    If not specified, ./tool_metadata.yml and ./dependencies/tool_metadata.yml
    are both checked if they exist.

    If none are specified, execution is halted.

    Otherwise, the resulting metadata file is parsed for an open_pdks version,
    which is then returned.
    """
    if version is not None:
        return version

    import yaml

    if tool_metadata_file_path is None:
        tool_metadata_file_path = os.path.join(".", "tool_metadata.yml")
        if not os.path.isfile(tool_metadata_file_path):
            tool_metadata_file_path = os.path.join(
                ".", "dependencies", "tool_metadata.yml"
            )
            if not os.path.isfile(tool_metadata_file_path):
                raise FileNotFoundError(
                    "Any of ./tool_metadata.yml or ./dependencies/tool_metadata.yml not found. You'll need to specify the file path or the commits explicitly."
                )

    tool_metadata = yaml.safe_load(open(tool_metadata_file_path).read())

    open_pdks_list = [tool for tool in tool_metadata if tool["name"] == "open_pdks"]

    if len(open_pdks_list) < 1:
        raise ValueError("No entry for open_pdks found in tool_metadata.yml")

    version = open_pdks_list[0]["commit"]

    return version


# -- The Deprecation Zone


def get_current_version(pdk_root: str, pdk: str) -> Optional[str]:
    # DEPRECATED: Use `Version.get_current()`
    return _get_current_version(pdk_root, pdk)


def get_installed_list(pdk_root: str, pdk: str) -> list:
    # DEPRECATED: Use `Version.get_all_installed()`
    return Version.get_all_installed(pdk_root, pdk)


def get_version_dir(pdk_root: str, pdk: str, version: Union[str, Version]) -> str:
    # DEPRECATED: Use `Version().get_dir()`
    if not isinstance(version, Version):
        version = Version(version, pdk)
    return version.get_dir(pdk_root)



# Copyright 2022-2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List, Dict


class Family(object):
    by_name: Dict[str, "Family"] = {}

    def __init__(
        self,
        name: str,
        variants: List[str],
        default_includes: List[str],
        all_libraries: List[str],
    ):
        self.name = name
        self.variants = variants
        self.default_includes = default_includes
        self.all_libraries = all_libraries


Family.by_name = {}
Family.by_name["sky130"] = Family(
    "sky130",
    ["sky130A", "sky130B"],
    [
        "sky130_fd_io",
        "sky130_fd_pr",
        "sky130_fd_pr_reram",
        "sky130_fd_sc_hd",
        "sky130_fd_sc_hvl",
        "sky130_ml_xx_hd",
        "sky130_sram_macros",
    ],
    [
        "sky130_fd_io",
        "sky130_fd_pr",
        "sky130_fd_pr_reram",
        "sky130_ml_xx_hd",
        "sky130_fd_sc_hd",
        "sky130_fd_sc_hdll",
        "sky130_fd_sc_lp",
        "sky130_fd_sc_hvl",
        "sky130_fd_sc_ls",
        "sky130_fd_sc_ms",
        "sky130_fd_sc_hs",
        "sky130_sram_macros",
    ],
)
Family.by_name["gf180mcu"] = Family(
    "gf180mcu",
    ["gf180mcuA", "gf180mcuB", "gf180mcuC", "gf180mcuD"],
    [
        "gf180mcu_fd_io",
        "gf180mcu_fd_pr",
        "gf180mcu_fd_sc_mcu7t5v0",
        "gf180mcu_fd_sc_mcu9t5v0",
        "gf180mcu_fd_ip_sram",
    ],
    [
        "gf180mcu_fd_io",
        "gf180mcu_fd_pr",
        "gf180mcu_fd_sc_mcu7t5v0",
        "gf180mcu_fd_sc_mcu9t5v0",
        "gf180mcu_fd_ip_sram",
        "gf180mcu_osu_sc_gp12t3v3",
        "gf180mcu_osu_sc_gp9t3v3",
    ],
)
Family.by_name["asap7"] = Family(
    "asap7", ["asap7"], default_includes=[], all_libraries=[]
)



# Copyright 2022 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import re
import shutil
import subprocess

from rich.progress import Progress

from ..common import mkdirp


class Repository(object):
    @classmethod
    def from_path(Self, path):
        name = os.path.basename(path)
        url = subprocess.check_output(
            ["git", "remote", "get-url", "origin"], stderr=subprocess.PIPE
        ).strip()

        remote_branch_info = open(
            os.path.join(path, ".git", "refs", "remotes", "origin", "HEAD")
        ).read()
        remote_branch = os.path.basename(remote_branch_info)

        return Self(name, url, path, remote_branch)

    def __init__(self, name, url, path, default_branch="main"):
        path = os.path.abspath(path)

        self.name = name
        self.url = url
        self.path = path
        self.default_branch = default_branch

    def clone_if_not_exist(self, callback=None):
        if os.path.exists(self.path):
            self.pristine()
            self.pull(callback)
        else:
            self.clone(callback)

    def clone(self, callback=None):
        try:
            shutil.rmtree(self.path)
        except FileNotFoundError:
            pass

        callback(0, f"Cloning {self.name}…")

        process = subprocess.Popen(
            ["git", "clone", "--progress", self.url, self.path],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        assert process.stderr is not None

        ro_rx = re.compile(r"Receiving objects:\s*(\d+)%")

        # Python Moment
        buffer = ""
        while True:
            bytes_read = process.stderr.read(1)
            if len(bytes_read) == 0:
                break
            char_read = bytes_read.decode("utf8")
            if char_read in ["\n", "\r"]:
                match = ro_rx.search(buffer)
                if match is not None:
                    if callback is not None:
                        callback(int(match[1]))
                buffer = ""
            else:
                buffer += char_read

        process.wait()

        subprocess.check_output(
            ["git", "submodule", "init"], stderr=subprocess.PIPE, cwd=self.path
        )

    def pristine(self):
        subprocess.check_output(
            ["git", "clean", "-fdX"], cwd=self.path, stderr=subprocess.PIPE
        )

        subprocess.check_output(
            ["git", "reset", "--hard", "HEAD"], cwd=self.path, stderr=subprocess.PIPE
        )

    def pull(self, callback=None):
        subprocess.check_output(
            ["git", "checkout", "-f", self.default_branch],
            cwd=self.path,
            stderr=subprocess.PIPE,
        )
        process = subprocess.Popen(
            ["git", "pull", "--no-recurse-submodules", "--progress"],
            cwd=self.path,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        assert process.stderr is not None
        callback(0, f"Updating {self.name}…")

        ro_rx = re.compile(r"Receiving objects:\s*(\d+)%")

        # Python Moment #2
        buffer = ""
        while True:
            bytes_read = process.stderr.read(1)
            if len(bytes_read) == 0:
                break
            char_read = bytes_read.decode("utf8")
            if char_read in ["\n", "\r"]:
                match = ro_rx.search(buffer)
                if match is not None:
                    if callback is not None:
                        callback(int(match[1]))
                buffer = ""
            else:
                buffer += char_read

        process.wait()
        if callback is not None:
            callback(100)
        subprocess.check_output(
            ["git", "submodule", "init"], stderr=subprocess.PIPE, cwd=self.path
        )

    def checkout_commit(self, commit: str):
        subprocess.check_output(
            ["git", "checkout", "-f", self.default_branch],
            cwd=self.path,
            stderr=subprocess.PIPE,
        )
        try:
            subprocess.check_output(
                ["git", "branch", "-f", "-D", "current"],
                cwd=self.path,
                stderr=subprocess.PIPE,
            )
        except Exception:
            pass
        subprocess.check_output(
            ["git", "checkout", "-f", "-b", "current", commit],
            cwd=self.path,
            stderr=subprocess.PIPE,
        )

    def init_submodule_if_not_exist(self, submodule: str, callback=None):
        submodule_git_path = os.path.join(self.path, submodule, ".git")
        if os.path.exists(submodule_git_path):
            subprocess.check_output(
                ["git", "submodule", "update", "--remote", submodule],
                cwd=self.path,
                stderr=subprocess.PIPE,
            )
            if callback is not None:
                callback(100)
        else:
            self.init_submodule(submodule, callback)

    def init_submodule(self, submodule: str, callback=None):
        process = subprocess.Popen(
            ["git", "submodule", "update", "--progress", submodule, submodule],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=self.path,
        )

        ro_rx = re.compile(r"Receiving objects:\s*(\d+)%")

        assert process.stderr is not None, "Process doesn't have a stderr channel"

        # Python Moment #3
        buffer = ""
        while True:
            bytes_read = process.stderr.read(1)
            if len(bytes_read) == 0:
                break
            char_read = bytes_read.decode("utf8")
            if char_read in ["\n", "\r"]:
                match = ro_rx.search(buffer)
                if match is not None:
                    if callback is not None:
                        callback(int(match[1]))
                buffer = ""
            else:
                buffer += char_read

        process.wait()


class GitMultiClone(object):
    progress: Progress

    def __init__(self, folder, progress):
        super().__init__()
        self.folder = os.path.abspath(folder)
        mkdirp(self.folder)
        self.progress = progress

    def clone(
        self, repo_url: str, commit: str, default_branch: str = "main"
    ) -> Repository:
        current_task = self.progress.add_task("", total=100)
        name = os.path.basename(repo_url)
        path = os.path.join(self.folder, name)
        r = Repository(name, repo_url, path, default_branch=default_branch)
        r.clone_if_not_exist(
            lambda x, y=None: self.progress.update(
                current_task, completed=x, description=y
            )
        )
        r.checkout_commit(commit)
        return r

    def clone_submodule(self, repo: Repository, submodule: str):
        current_task = self.progress.add_task(
            f"Updating submodule {submodule}…", total=100
        )
        repo.init_submodule_if_not_exist(
            submodule, lambda x: self.progress.update(current_task, completed=x)
        )



# Copyright 2022-2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import shutil
import subprocess
from typing import Optional, List, Dict
from concurrent.futures import ThreadPoolExecutor

from rich.console import Console
from rich.progress import Progress

from .common import RepoMetadata
from .git_multi_clone import GitMultiClone
from ..common import (
    get_version_dir,
    get_volare_dir,
    mkdirp,
)
from ..families import Family

repo_metadata = {
    "orfs": RepoMetadata(
        "https://github.com/The-OpenROAD-Project/OpenROAD-flow-scripts",
        "b541139f4da34687340bf2e55528707e76fce233",
        "master",
    )
}


def get_orfs(version, build_directory, jobs=1):
    try:
        console = Console()

        orfs_repo = None

        with Progress() as progress:
            with ThreadPoolExecutor(max_workers=jobs) as executor:
                gmc = GitMultiClone(build_directory, progress)
                orfs = repo_metadata["orfs"]
                orfs_future = executor.submit(
                    GitMultiClone.clone,
                    gmc,
                    orfs.repo,
                    version,
                    orfs.default_branch,
                )
                orfs_repo = orfs_future.result()

        console.log(f"Done fetching {orfs_repo.name}.")

    except subprocess.CalledProcessError as e:
        print(e)
        print(e.stderr)
        exit(-1)


def build_variants(build_directory, jobs):
    PDK_LINKER_SCRIPT = "https://raw.githubusercontent.com/The-OpenROAD-Project/OpenLane/01e951092150ee8619286b0807ee263198b5ea6d/scripts/pdk-linker.py"

    subprocess.check_call(
        [
            "sed",
            "-i",
            "s/grid_strategy-M2-M5-M7.cfg/grid_strategy-M2-M5-M7.tcl/",
            "./OpenROAD-flow-scripts/flow/platforms/asap7/openlane/mapping.json",
        ],
        cwd=build_directory,
    )
    subprocess.check_call(
        [
            "sed",
            "-i",
            "s/grid_strategy-M2-M5-M7.cfg/grid_strategy-M2-M5-M7.tcl/",
            "./OpenROAD-flow-scripts/flow/platforms/asap7/openlane/asap7sc7p5t/config.tcl",
        ],
        cwd=build_directory,
    )

    subprocess.check_call(["curl", "-LO", PDK_LINKER_SCRIPT], cwd=build_directory)

    subprocess.check_call(
        [
            "python3",
            "./pdk-linker.py",
            "-s",
            "OpenROAD-flow-scripts",
            "-d",
            "asap7",
            "-m",
            "./OpenROAD-flow-scripts/flow/platforms/asap7/openlane/mapping.json",
        ],
        cwd=build_directory,
    )


def install_asap7(build_directory, pdk_root, version):
    console = Console()
    with console.status("Adding build to list of installed versions…"):
        version_directory = get_version_dir(pdk_root, "asap7", version)
        print(version_directory)
        if (
            os.path.exists(version_directory)
            and len(os.listdir(version_directory)) != 0
        ):
            backup_path = version_directory
            it = 0
            while os.path.exists(backup_path) and len(os.listdir(backup_path)) != 0:
                it += 1
                backup_path = get_version_dir(pdk_root, "asap7", f"{version}.bk{it}")
            console.log(
                f"Build already found at {version_directory}, moving to {backup_path}…"
            )
            shutil.move(version_directory, backup_path)

        console.log("Copying…")
        mkdirp(version_directory)

        asap7_family = Family.by_name["asap7"]

        for variant in asap7_family.variants:
            variant_build_path = os.path.join(build_directory, variant)
            variant_install_path = os.path.join(version_directory, variant)
            if os.path.isdir(variant_build_path):
                shutil.copytree(variant_build_path, variant_install_path)

    console.log("Done.")


def build(
    pdk_root: str,
    version: str,
    jobs: int = 1,
    clear_build_artifacts: bool = True,
    include_libraries: Optional[List[str]] = None,
    using_repos: Optional[Dict[str, str]] = None,
    build_magic: bool = False,
):
    if using_repos is None:
        using_repos = {}

    # TODO: Support using_repos
    build_directory = os.path.join(get_volare_dir(pdk_root, "asap7"), "build", version)

    get_orfs(version, build_directory, jobs)
    build_variants(build_directory, jobs)
    install_asap7(build_directory, pdk_root, version)



#!/usr/bin/env python3
import os
import subprocess
from setuptools import setup, find_packages

module_name = "volare"

__dir__ = os.path.abspath(os.path.dirname(__file__))
version = subprocess.check_output(
    [
        "python3",
        os.path.join(
            __dir__,
            module_name,
            "__version__.py",
        ),
    ],
    encoding="utf8",
)

requirements = (
    open(os.path.join(__dir__, "requirements.txt")).read().strip().split("\n")
)

setup(
    name=module_name,
    packages=find_packages(),
    package_data={"volare": ["py.typed"]},
    version=version,
    description="An open_pdks PDK builder/version manager",
    long_description=open(os.path.join(__dir__, "Readme.md")).read(),
    long_description_content_type="text/markdown",
    author="Efabless Corporation",
    author_email="donn@efabless.com",
    install_requires=requirements,
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
    ],
    entry_points={"console_scripts": ["volare = volare.__main__:cli"]},
    python_requires=">3.6",
)
